#include <algorithm>
#include <array>
#include <chrono>
#include <condition_variable>
#include <iostream>
#include <mutex>
#include <thread>
#include <vector>

#define NUMBER_OF_GUYS 3
#define NUMBER_OF_TOOLS 3
#define WORK_ITERATIONS 100

using namespace std;
using namespace std::chrono;

mutex outputMtx;	// Mutex used to avoid garbled output.
mutex borrowOrReturnMtx;	// Mutex used for the following condition variable.
condition_variable condVar;	// Threads are notified if tools are returned.

const array<const char*, NUMBER_OF_TOOLS> toolsName = { "hammer", "screw driver", "saw" };
int numberOfTools[NUMBER_OF_TOOLS] = { 1, 1, 1 };	// Initially, just one tool from each.

enum Tools {
	HAMMER = 0, SCREW_DRIVER = 1, SAW = 2
};

void work() {
	// Heavy work of a guy, just a random delay...
	this_thread::sleep_for(chrono::milliseconds(rand() % 30));
}

void borrowTools(const vector<int>& toolsToBorrow)	{
	unique_lock<mutex> lock(borrowOrReturnMtx);
	condVar.wait(lock, [&toolsToBorrow]() {
		return all_of(toolsToBorrow.cbegin(), toolsToBorrow.cend(),
				[](int tool) { return numberOfTools[tool] > 0; });
	});

	/*
	 * If a thread is woken and the predicate is satisfied (all tools available),
	 * then mutex 'borrowOrReturnMtx' is still locked and we can modify the tool numbers.
	 */

	for (int toolIdx : toolsToBorrow)
		--numberOfTools[toolIdx];

	// After the method is finished, mutex 'borrowOrReturnMtx' will be unlocked.
}

void returnTools(const vector<int>& toolsToReturn)	{
	unique_lock<mutex> mutex_lock(borrowOrReturnMtx);

	// Mutex is locked, no other thread can return or borrow tools until it is unlocked.
	for (int toolIdx : toolsToReturn)
		++numberOfTools[toolIdx];

	// Notify that some tools are returned and try to awake all threads one by one.
	condVar.notify_all();
}

void guyJob(int threadId, int tool1, int tool2) {
	// Each guy is iteratively doing the job.
	for (uint32_t iter = 0; iter < WORK_ITERATIONS; ++iter)	{

		/* WORK WITH ONE TOOL */
		borrowTools({ tool1 });

		outputMtx.lock();
		cout<<"Guy "<<threadId<<" borrowed "<<toolsName[tool1]<<"."<<endl;
		outputMtx.unlock();

		work();	

		returnTools({ tool1 });

		outputMtx.lock();
		cout<<"Guy "<<threadId<<" returned "<<toolsName[tool1]<<"."<<endl;
		outputMtx.unlock();

		/* WORK WITH TWO TOOLS */

		borrowTools({ tool1, tool2 });

		outputMtx.lock();
		cout<<"Guy "<<threadId<<" borrowed "<<toolsName[tool1]<<" and "<<toolsName[tool2]<<"."<<endl;
		outputMtx.unlock();

		work();

		returnTools({ tool1, tool2 });

		outputMtx.lock();
		cout<<"Guy "<<threadId<<" returned "<<toolsName[tool1]<<" and "<<toolsName[tool2]<<"."<<endl;
		outputMtx.unlock();
	}

	return;
}

int main() {
	// Assign a pair of tools to each guy.
	vector<pair<int, int>> guyTools = {
		{HAMMER, SCREW_DRIVER}, {SCREW_DRIVER, SAW}, {SAW, HAMMER}
	};

	// Let every guy to work with the assigned tools.
	vector<thread> guys;
	for (int g = 0; g < NUMBER_OF_GUYS; ++g)
		guys.emplace_back(guyJob, g, guyTools[g].first, guyTools[g].second);
		
	// Wait until the guys finish their work.
	for (thread& th : guys)
		th.join();

	return 0;
}

